
/** INCLUDES *******************************************************/

#include <p18f4550.h>
#include <timers.h>
#include <adc.h>
#include "sensors.h"        

// INTOSC => 8 MHz
#pragma config FOSC=INTOSC_XT

#pragma config FCMEN=OFF
#pragma config IESO=OFF
#pragma config VREGEN=ON
#pragma config CCP2MX=OFF
#pragma config PBADEN=OFF

#pragma config PWRT=ON
#pragma config MCLRE=ON
#pragma config BOR=ON		
#pragma config BORV=2

//#pragma config WDT=ON
#pragma config WDT=OFF
//#pragma config WDTPS=8192
	
#pragma config STVREN=ON
#pragma config LVP=OFF
#pragma config ICPRT= OFF
#pragma config XINST=OFF
#pragma config DEBUG=OFF
#pragma config WRTB=ON
#pragma config LPT1OSC=OFF

/** VARIABLES ******************************************************/
#pragma udata



#define LED_ON   1u
#define LED_OFF  0u

#define POWER_ON 0u
#define POWER_OFF 1u

#define START_PRESSED 0u

#define RELAY_ON   1u
#define RELAY_OFF  0u

#define CONT_ON   0u
#define CONT_OFF  1u

#define NO2C_ON   1u
#define NO2C_OFF  0u

// define output ports

#define P_LED_ON     PORTBbits.RB5
#define P_LED_READY  PORTBbits.RB4
#define P_SW1        PORTBbits.RB3
#define P_NO2C       PORTBbits.RB2
#define P_SW2        PORTDbits.RD6
#define P_SW3        PORTDbits.RD4

// enable output ports

#define TRIS_OUTPUT TRISBbits.TRISB5=0;  \
                    TRISBbits.TRISB4=0;  \
                    TRISBbits.TRISB3=0;  \
                    TRISBbits.TRISB2=0;  \
                    TRISDbits.TRISD6=0;  \
                    TRISDbits.TRISD4=0;



// define input ports
#define P_CONT       PORTBbits.RB1
#define P_BYPASS     PORTBbits.RB0
#define P_POWER      PORTDbits.RD7
#define P_START      PORTDbits.RD5

// enable input ports

#define TRIS_INPUT TRISBbits.TRISB1=1; \
                   TRISBbits.TRISB0=1; \
                   TRISDbits.TRISD7=1; \
                   TRISDbits.TRISD5=1;


volatile int power_off_delay = 0;

#define POWER_OFF_DELAY_SET(VALUE)  power_off_delay = VALUE
#define POWER_OFF_DELAY_FINISHED   (power_off_delay>0)?0:1
#define POWER_OFF_DELAY_DEC         power_off_delay--

volatile int delay = 0;

#define DELAY_SET(VALUE)  delay = VALUE
#define DELAY_FINISHED   ((delay>0)?0:1)
#define DELAY_DEC         delay--

/*
#define DELAY_SECOND      8
#define DELAY_SECOND_4    2
*/

                /* why does the 20*DELAY_SECOND not work? */

#define TIME_SECOND             8 
#define TIME_POWER_ON          80 /* 10*DELAY_SECOND */
#define TIME_BUTTON_PRESS       2 /* DELAY_SECOND_4 */
#define TIME_BYPASS            48 /* 6*DELAY_SECOND */
#define TIME_TIMEOUT          160 /* 20*DELAY_SECOND */
#define TIME_POWER_OFF          4 /* 2* DELAY_SECOND_4 */


// STATES
#define ST_BEGIN          0u                   
#define ST_OFF            1u
#define ST_ON             2u
#define ST_ON_WAIT        3u
#define ST_START          4u
#define ST_START_PRESSED  5u
#define ST_WAIT_READY     6u
#define ST_READY          7u
#define ST_PRESS_START    8u
#define ST_BYPASS_WAIT    9u
#define ST_CONT_WAIT     10u
#define ST_RUNNING       11u

int state = ST_BEGIN;
 
 
extern unsigned int VALUE_VSOURCE;

// VSOURCE_LIMIT * 5 / 1024 = Volts
#define VSOURCE_MIN  614u  

/** PRIVATE PROTOTYPES *********************************************/

void LowISR(void);
void HighISR(void);

#pragma code

#pragma interruptlow LowISR
void LowISR(void)
  {
  
  }

#pragma interruptlow HighISR
void HighISR(void)
  {	
  
  	//Check which interrupt flag caused the interrupt.
  	if (INTCONbits.TMR0IF)
  	  {
  	  /* The TMR1 register pair (TMR1H:TMR1L) increments
        * from 0000h to FFFFh and rolls over to 0000h. The
        * Timer1 interrupt, if enabled, is generated on overflow,
        * which is latched in interrupt flag bit, TMR1IF
        * (PIR1<0>). This interrupt can be enabled or disabled
        * by setting or clearing the Timer1 Interrupt Enable bit,
        * TMR1IE (PIE1<0>). 
        */

        
   		INTCONbits.TMR0IF = 0;     // clear the interrupt
   		
   		DELAY_DEC;  // decrement delay value
   		POWER_OFF_DELAY_DEC;
   		
  	}
  }

#pragma code lowVector=0x18
void LowVector(void){_asm goto LowISR _endasm}

#pragma code highVector=0x8
void HighVector(void){_asm goto HighISR _endasm}

#pragma code  // Return to default code section



void main(void)
  {   
    unsigned int st_state = ST_BEGIN;
    unsigned int shutdown = 0;
    
    ADCON1 |= 0x0F;                 // Default all pins to digital
    
    // ensure 8MHz clock
    OSCCONbits.IRCF2 = 1;
    OSCCONbits.IRCF1 = 1;
    OSCCONbits.IRCF0 = 1;
    
    P_LED_ON = LED_OFF;
    P_LED_READY = LED_OFF;
    
    P_NO2C = NO2C_OFF;
    P_SW2 = RELAY_OFF;
    P_SW1 = RELAY_OFF;
    P_SW3 = RELAY_OFF;
         
    // enable output pins
    TRIS_OUTPUT;
    
    // enable input pins
    TRIS_INPUT;

    // initialize sensors 
    sensors_init();
    
    // initialize timer
    // http://www.camotruck.net/rollins/electronics/pic_timer.html
    OpenTimer0( 
        TIMER_INT_ON & 
        T0_SOURCE_INT &
        T0_16BIT &
        T0_PS_1_4  // 1 => 0.13 seconds
      ); 
          
    // INTCON=0;            //make sure interrupts are disable
    // INTCONbits.GIE=1;    //enable global interrupts
    // INTCONbits.PEIE=1;   //enable peripheral interrupts
    // INTCONbits.TMR0IE=1; //enable TMR0 overflow interrupt enable bit
    INTCON=0xe0;            //all 4 lines above combined
       
       
    while(1)
      {
        
        sensors_update();
        
        // check specific conditions
        // is the power still on ?
        // if (P_POWER_ON)
        // if (P_BYPASS)
        
        if (!(st_state == ST_BEGIN))
          {
            // as the system looses power, the P_POWER and P_BYPASS will be OFF but it will be 
            // in the ST_BEGIN state.  We don't need to run this again.
            
            if ((P_POWER == POWER_OFF) && (P_BYPASS == POWER_OFF))
              {
                // Both power feeds are off.  
                // This happens when the key is turned off or for a possible instant when the relays swap.
                
                // the first time "shutdown" will be 0
                if (shutdown)
                  {
                    if (POWER_OFF_DELAY_FINISHED)
                      {
                        // we still have no power and the timer has expired.  
                        // Let's go back to the BEGIN state and wait to turn off.
                        st_state = ST_BEGIN;  
                        shutdown = 0;                 
                      }
                  }
                else
                  {
                    // set shutdown to 1 and start the timer.
                    shutdown = 1; 
                    POWER_OFF_DELAY_SET( TIME_POWER_OFF ); 
                  }
              }
            else
              {
                // if power is returned to either input then go out of shutdown mode.
                shutdown = 0;
              }
          }

          
     		      
        if (VALUE_VSOURCE < VSOURCE_MIN) 
          {
             // This means we are running out of +12 power.
             // shut everyting off and wait to die.
             st_state = ST_BEGIN;   
          }
          
        
                  
     		switch(st_state)
     		  {
       		  
       		  case ST_BEGIN:       		          
       		      
       		      P_LED_READY = LED_OFF;
       		      P_LED_ON = LED_OFF; 
 
                P_SW2 = RELAY_OFF;
                
                DELAY_SET( TIME_SECOND );  
                while(! DELAY_FINISHED );
                
                P_SW1 = RELAY_OFF;
                P_SW3 = RELAY_OFF;
                // P_NO2C = NO2C_OFF;      // we don't need to turn this off every time.
                   		             		   
       		      // check to see if the source voltage is over the minimum value
       		            		      
                if (VALUE_VSOURCE > VSOURCE_MIN) 
                  {
                    P_NO2C = NO2C_ON;  
                    st_state = ST_ON;   
                  }                 
              break;
              
       		  case ST_ON:
       		      if (P_POWER == POWER_ON)
       		        {
         		        shutdown = 0;
         		        st_state = ST_ON_WAIT;
         		        DELAY_SET( TIME_POWER_ON );
         		        //DELAY_SET( DELAY_POWER_ON );
         		      }
              break; 
              
            case ST_ON_WAIT:
       		      if (P_POWER == POWER_ON)
       		        {
         		        if (DELAY_FINISHED) 
         		          st_state = ST_READY;
         		      }
         		    else
                  st_state = ST_ON; // key on and then key off before delay expired            
              break;
              
       		  case ST_READY:
       		      P_LED_READY = LED_ON;

                if (P_START == START_PRESSED)
                  {
                    // the start button was pressed
                    st_state = ST_START_PRESSED;
                    
                    // check to see if the button is pressed for 1/4 second
                    DELAY_SET( TIME_BUTTON_PRESS );                    
                    // DELAY_SET( DELAY_SECOND_4 );
                  }
              break;

            case ST_START_PRESSED:
                if (P_START == START_PRESSED)
                  {
                     if (DELAY_FINISHED) 
                       {
         		             st_state = ST_START;
         		             P_LED_ON = LED_ON;
                   		             
                         DELAY_SET( TIME_SECOND );  
                         while(! DELAY_FINISHED );
         		           }
                  }   
                else
                  st_state = ST_READY;
              break;    
                   
  
       		  case ST_START:
       		      // enter into bypass mode and remain there for a few seconds
       		      
       		      // Engage the SW1 relay
       		      // - disable +12 to all components
       		      // - send +144 to the controller
       		      P_SW1 = RELAY_ON;
       		      
       		      // enter a wait state to ensure the DC-DC converter is stable and
       		      // allow the +144 to feed the controller
       		      st_state = ST_BYPASS_WAIT;
       		      
                DELAY_SET( TIME_BYPASS );       
                
              break;
              
            case ST_BYPASS_WAIT:
                if (DELAY_FINISHED) 
                  {        
                    P_LED_READY = LED_OFF;
                             		             
                    // after waiting in bypass mode engage the relay to send power
                    // to the controller and the NO1C+
                    P_SW2 = RELAY_ON;
                    
                    // enter the CONT_WAIT state and check for the CONT to be engaged
                    st_state = ST_CONT_WAIT;  
                    
                    // set the delay to timeout if it takes too long
                    DELAY_SET( TIME_TIMEOUT );                  
                  }            
              break;
                   

            case ST_CONT_WAIT:
                if (P_CONT == CONT_ON) 
                  { 
                    // The system is ready
                    // Turn on the relay to exit bypass mode.  
                    P_SW3 = RELAY_ON;

         		        st_state = ST_RUNNING;
                  }
                else if (DELAY_FINISHED) 
                  {
                    // we have a problem.  The CONT_WAIT is taking too long.
                    // go back to the begin state
                    st_state = ST_BEGIN;                  
                  }                  
              break;                   

            case ST_RUNNING:
                // the state transitions from here will be executed outside of this
                // switch statement. 
              break;          
                                                     
         } // switch       
                                  
        
      }
  }

